import emerge as em
from emerge.plot import plot_sp, plot_ff, plot_ff_polar, smith
import numpy as np

""" DEMO 18: Plotting and Visualization

In this demo we will show the various plotting and visualization features in EMerge. 
We will use a quite simple geometry with some basic features that can run quickly to showcase
all the features in EMerge. 

The model will be a simple WR90 waveguide with a small inductive iris radiating into a half sphere.
In front of the radiator we place a 5mm radius copper ball, just for visualization purposes, to make things interesting!

It is not supposed to do anything special nor be well matched. Its only about demonstrating plot features.

"""


# First define some constants

mm = 0.001
wga = 22.86*mm
wgb = 10.16*mm
L = 50*mm
rad = 30*mm
th = 1*mm
opening = 0.8
distance = 2*mm

# Setup the simulation

m = em.Simulation('PlottingDemo')

box = em.geo.Box(wgb, wga, L, (-wgb/2, -wga/2, -L))
cut1 = em.geo.Box(wgb, wga*(1-opening)/2, th, (-wgb/2, -wga/2, -distance-th/2))
cut2 = em.geo.Box(wgb, wga*(1-opening)/2, th, (-wgb/2, wga*opening/2, -distance-th/2))

feed = em.geo.subtract(box, cut1)
feed = em.geo.subtract(feed, cut2)

halfsphere = em.geo.HalfSphere(rad, direction=(0,0,1))

sphere = em.geo.Sphere(5*mm, (0,0,10*mm)).set_material(em.lib.COPPER) # This sphere is just for instersesting visuals and scattering.
sphere.max_meshsize = 1*mm

# The simplest way to view the geometry is to call the "view" method on our simulation object.
# The view method can use the GMSH viewer or the EMerge PyVista viewer. 
# As long as there is no mesh yet, EMerge can only use the GMSH viewer.

#m.view()

# We commit the geometry and generate the mesh.
m.commit_geometry()
m.mw.set_frequency_range(9e9, 10e9, 3)
m.mw.set_resolution(0.2)
m.generate_mesh()

# If we now call view, we will see the PyVista plot window. Just click it away if you are done to proceed to the next.
m.view()

# We can pass one or more selections to the viewer to highlight them in red.

m.view(selections=[feed.face('bottom', tool=box)])

# We can also view all object labels by turning labels on

m.view(labels=True)

# Whenever it is important to know face names for selections we can simply show all of them like this for a specific geometry.

m.view(selections=feed.all_faces())

# or this for all faces

m.view(face_labels=True)

# We can also view the mesh in various ways

m.view(plot_mesh=True)
m.view(plot_mesh=True, volume_mesh=False) # Faces only!

############################################################
#                        SIMULATION                       #
############################################################

# Now lets run our simulation to view the data we produce.

abc = halfsphere.outside
my_port = m.mw.bc.RectangularWaveguide(feed.face('-z'), 1)
m.mw.bc.AbsorbingBoundary(abc)

data = m.mw.run_sweep()

glob = data.scalar.grid

############################################################
#                         1D PLOTS                        #
############################################################

# We have now created our data. First lets look at some 1D plot options. First we can visualize S-parameters

plot_sp(glob.freq, glob.S(1,1), labels=['S11',])

# Or a smith plot

smith(glob.S(1,1), f=glob.freq) # The frequency is not necessary but sometimes useful to show!


############################################################
#                      FARIELD PLOTS                      #
############################################################

# There are also some useful farfield plot functions. First lets compute the farfield in 2 slices

field = data.field.find(freq=9.5e9)

ff_azi = field.farfield_2d((0,0,1), (0,1,0), abc, (-90, 90))
ff_ele = field.farfield_2d((0,0,1), (1,0,0), abc, (-90, 90))

plot_ff(ff_azi.ang*180/3.1415, [ff_azi.normE/em.EISO, ff_ele.normE/em.EISO], dB=True, labels=['E-plane','H-plane'], ylabel='Gain [dBi]')

# Or a polar plot

plot_ff_polar(ff_azi.ang, [ff_azi.normE/em.EISO, ff_ele.normE/em.EISO], dB=True, labels=['E-plane','H-plane'])


############################################################
#                         3D PLOTS                        #
############################################################

# We will now look at some of the 3D plot functionality.
# To customize 3D plot windows, we use the display attribute of our simulation model.
# We can name it for simplicity

dp = m.display

# I will now go through some steps to demonstrate how it works.
# By default the Display is empty, we have to populate it.

dp.show()

# We can add objects one by one...

dp.add_object(feed)
dp.add_object(halfsphere)
dp.add_object(sphere)

dp.show()

# Or in sequence

dp.add_objects(feed, halfsphere, sphere)
dp.show()

# Or automatically

dp.add_objects(*m.all_geos())
dp.show()

# The default add object method has some extra features to change how geometries are visulaized like opacity.

dp.add_object(feed)
dp.add_object(halfsphere)
dp.add_object(sphere, opacity=0.2)

dp.show()

""" We often want to add fields to visualize. We will add those now.
The display class has no direct features to add cut-plane E-fields or stuff like that. Instead it just has a
surface plot function .add_surf(X,Y,Z,Scalar) that takes at least 4 positional arguments. 
    X: 2D array of X-coordinates
    Y: 2D array of Y-coordinates
    Z: 2D array of Z-coordinates
    Scalars: 2C array of scalar values
    
Thus we need to create those. To create that data we can use our dataset. By using the .find() method we create
a Microwave field object that we can use to interpolate our E-field. The slow way to create a good dataset for a cutplane
is as following.
"""
ys = np.arange(-rad, rad, 2*mm)
zs = np.arange(-L, rad, 2*mm)
Y,Z = np.meshgrid(ys, zs)
X = 0*Y
field_data = field.interpolate(X,Y,Z)

dp.add_objects(*m.all_geos())
dp.add_surf(X,Y,Z,field_data.Ex.real, symmetrize=True) # Symmetrize automatically makes the color range symemtrical and selects a cool color map.
dp.show()

# Of course we don't want to go through this laboreous process every time. EMerge has some convenient functions in the MWData 
# class that makes creating this data and the plot arguments easier.

dp.add_objects(*m.all_geos())
dp.add_surf(*field.cutplane(1*mm, x=0).scalar('Ex','real'), symmetrize=True)
dp.show()

# The .scalar() method automatrically returns X, Y, Z, and Scalar. the * notation unpacks that. 
# To animate this we simply add the animate call.

#                                                               v Don't forget to set this to complex!
dp.add_objects(*m.all_geos())
dp.animate().add_surf(*field.cutplane(1*mm, x=0).scalar('Ex','complex'), symmetrize=True)
dp.show()

# Plotting fields on the surface of an object is also possible.
dp.add_objects(*m.all_geos())
dp.add_boundary_field(abc, field.boundary(abc).scalar('Ex','abs'), clim=(0, 600)) # we don't unpack here because of how this method is designed.
dp.show()

# To ensure that two color bars are shared we can use the .cbar function

dp.add_objects(*m.all_geos())
dp.cbar('normE [V/m]', clim=(0, 4000)).add_surf(*field.cutplane(1*mm, x=0).scalar('normE'), symmetrize=False)
dp.cbar('normE [V/m]').add_surf(*field.cutplane(1*mm, y=0).scalar('normE'), symmetrize=False)
dp.show()

# We can also add farfield patterns using .add_surf(). All we need to do is create a Farfield dataset using the farfield_3d method.
# This dataset has the .surfplot() method that generates the data needed to create a farfield bulp. In this method you can tune 

# isotropic (bool): If the isotropic radiated power should be computed.
# dB (bool): If the data should be computed in dB
# dBfloor (float): Which dB value should be considered 0mm radius. (farfield plots can otherwise go to -infinity dB)
# rmax (float): The radius corresponding to the largest Farfield value
# offset (float, float, float): The x,y,z coordinate at which the blob is centered.

ff3d = field.farfield_3d(abc)

dp.add_objects(*m.all_geos())
dp.add_surf(*field.cutplane(1*mm, x=0).scalar('Ex','real'), symmetrize=True)
dp.add_surf(*ff3d.surfplot('normE','abs',True, rmax=20*mm, offset=(0,0,30*mm)))
dp.show()

# Other useful methods are

griddata = field.grid(3.3*mm)
dp.add_objects(*m.all_geos())
dp.add_title('Some text here!')
dp.add_text('Instruction here','blue','lower_right')
dp.add_portmode(my_port, k0=field.k0) # k0 is not needed here but it is often with modal-ports.
dp.add_quiver(*griddata.vector('H','real'))
dp.add_contour(*griddata.scalar('Ex','real'), symmetrize=True, Nlevels=30)
dp.add_surf(*field.cutplane(1*mm, x=0).scalar('Ex','real'), symmetrize=True)
dp.add_surf(*ff3d.surfplot('normE','abs',True, rmax=20*mm, offset=(0,0,30*mm)))
dp.show()



# -----------------------------------------------------------------------------
# Microstrip to slotline coupler and Adaptive mesh refinement
#
# In this example we will look at the setup of a microstripline to slotline coupler
# and see how we can use Adaptive mesh refinement in this simulation
# -----------------------------------------------------------------------------


import emerge as em
import numpy as np
from emerge.plot import plot_sp


############################################################
#                    CONSTANT DEFINITION                   #
############################################################


mm = 0.001      # mm definition

th = 1.54       # PCB Thickness
Lprestub = 2    # Extra stripline before the radial stub
Lstub = 8       # Length of the radial stub
angstub = 100   # Angle of the radial stub
Lfeed = 20      # Length of the microstrip feedline
Lslot = 20      # Length of the slotline
wslot = 0.8     # width of the slotline
Lleft = 4       # Extra distance before the open-circuit circle
Rcirc = 8       # Readius of the open circuit circle.


############################################################
#                     SIMULATION SETUP                    #
############################################################

# We will invoke the SimulationBeta class because it houses some specific
# implementation details required for adaptive mesh refinement.
m = em.SimulationBeta('Transition', loglevel='INFO')
m.check_version("1.2.0")
# Next we create the PCB designer class instance.
pcb = em.geo.PCB(th, mm, material=em.lib.DIEL_RO4003C)

# We can use this function to compute the microstripline impedance. Right now (version 1.1) it just assumes top vs bottom layer.
w0 = pcb.calc.z0(50)

# We start a new simple microstripline at xy=(0,0) and go Lfeed forward.
pcb.new(0,0, w0, (0,1)).store('p1').straight(Lfeed).store('stop').straight(Lprestub).store('stub')

# We load the StripLine element from the PCB where we stopped and use the coordinates to add a radial stub.
stub = pcb.load('stub')
pcb.radial_stub(stub.xy, Lstub, angstub, (0,1), w0=w0)

# Then we use a botom layer stripline to create the slotline trace. We start a bit left of the stop point where
# and then proceed to the right.
stop = pcb.load('stop')
pcb.new(stop.x-Lleft, stop.y, wslot, (1,0), pcb.z(0)).store('circ').straight(Lslot+Lleft).store('p2')

# We load the XY coordinates of that last point
xydisc = pcb.load('circ').xy

# We create a circle geometry Such that the right side contacts the trace with 1mm overlap.
disc = em.geo.XYPolygon.circle(Rcirc*mm, 1*mm).geo(em.GCS.displace((xydisc[0]-Rcirc+1)*mm, xydisc[1]*mm, pcb.z(0)*mm))

# Then finally we generate all our paths. 
# They will be returned in the following order
#    1. All stripline paths in order of creation
#    2. All polygon geometries

feed, slot, stub = pcb.compile_paths()

# After generation of all geometries we can determine the bounds. We will add 20mm of margin to the left and back of
# our PCB domain. We start at the front and end at the right.
pcb.determine_bounds(20,20, 0, 0)

# We add a ground plane on the bottom.
ground = pcb.plane(pcb.z(0))

# We remove the slot line and circle from the ground plane
em.geo.subtract(ground, em.geo.unite(slot, disc))

# We add two modal port surfaces at the nodes p1 and p2
mp1 = pcb.modal_port(pcb.load('p1'), 5.0)
mp2 = pcb.modal_port(pcb.load('p2'), (5.0, 5.0), width_multiplier=15) #Wthe slot port width is 15 times the slot width.

# Finally we generate the PCB, top air and bottom air.
diel = pcb.generate_pcb()
air_top = pcb.generate_air(5)
air_bottom = pcb.generate_air(5, bottom=True)

# With Commit geometry we say: here we are done generating our model.
m.commit_geometry()

# We set the frequency range from 4GHz to 7GHz in 21 steps.
m.mw.set_frequency_range(4e9, 7e9, 21)

# We don't use any manual refinement steps to illustrate the power of
# Adaptive Mesh refinement.
# We generate the mesh and view it
m.mw.set_resolution(0.2)
m.generate_mesh()

# Notice the course initial mesh
m.view(plot_mesh=False)
m.view(plot_mesh=True)

############################################################
#                    BOUNDARY CONDITIONS                   #
############################################################

# Here we define our boundary conditions. Our first port is a (quasi-)TEM line. 
p1 = m.mw.bc.ModalPort(mp1, 1, modetype='TEM')
# A pure slotline is also (quasi-)TEM but because its surrounded by a PEC boundary it
# Behaves like a TE mode. We set this explicitly to ensure proper extrapolation of the out-of-plane
# propagation constant
p2 = m.mw.bc.ModalPort(mp2, 2, modetype='TE')
# We align the slotline mode along the positive Y-direction. This ensures that no polarity flips
# occur for the slotline mode that will add a random 180 degree phase shift in between
# refinement passes. The port phase is a degree of freedom that is not solved for during
# The modal port analysis so  this way we prevent that degree of freedom from showing up in 
# S-parameter convergence study.

p2.align_modes(em.YAX)

# we also add an absorbing boundary at the top and bottom face to minimize resonance modes
m.mw.bc.AbsorbingBoundary(em.select(air_top.top, air_bottom.bottom))
############################################################
#                        SIMULATION                       #
############################################################

# First we call the Adaptive Mesh Refinement routine. We refine at 6GHz because this is in the pass-band.
# We set show_mesh to True so we can see the progress of refinement for the purspose of this example.
# This halts the simulation so we have to click away the window to proceed.
# You can see that more nodes are added around the signal traces because the E-field error is highest
m.adaptive_mesh_refinement(frequency=5.5e9, show_mesh=True)

# We can view the improvement in the refined mesh.
m.view(plot_mesh=True, volume_mesh=False)

# Finally we start our solve with 4 parallel workers
data = m.mw.run_sweep(True, n_workers=4)


############################################################
#                      POST PROCESSING                     #
############################################################

# We make a convenient object name for the gridded S-parameter data and plot it.
g = data.scalar.grid

plot_sp(g.freq, [g.S(1,1), g.S(2,1)], labels=['S11','S21'])

# Finally we make a 3D plot showing an animation of the Ez-field.
field = data.field.find(freq=6e9)
m.display.add_objects(*m.all_geos())
m.display.animate().add_surf(*field.cutplane(1*mm, z=-th*mm/2).scalar('Ez','complex'), symmetrize=True)
m.display.add_portmode(p1, k0=field.k0)
m.display.add_portmode(p2, k0=field.k0)
m.display.show()




# -----------------------------------------------------------------------------
# STEP File import
#
# In this example we will look at how the STEP file import module can be used.
# Please read the following CAREFULLY
# -----------------------------------------------------------------------------

############################################################
#                         IMPORTANT                        #
############################################################
"""
This example file demonstrates how to import and use STEP files in EMerge.
Try toggling the if-statements on/off to explore behavior and meshing issues.
If the simulation halts, check if any display windows are hidden behind others.
"""

import emerge as em
from emerge.plot import plot_sp
from pathlib import Path

mm = 0.001  # unit helper: 1 mm in meters

margin = 5*mm
model = em.Simulation('STEP_Board_demo')  # create simulation container
model.check_version('1.2.0') # Checks version compatibility.

# STEPItems is a wrapper that loads geometries from a STEP file.
# It can contain multiple solids, surfaces, edges, and points.
# These are accessible via .volumes, .faces, .edges, and .points.
step = em.geo.STEPItems('board', str(Path(__file__).parent / 'Connector_MicrostripBoard_Solder.step'), 0.001)

# The imported STEP geometry can include multiple disconnected volumes.
# Visual inspection helps identify which volume represents which component.
# You can use m.generate_mesh() and m.view() to preview.
# Large cylindrical or thin features may cause meshing errors if element size is too coarse.

# SET TRUE TO LET IT CRASH
# Below block intentionally crashes to demonstrate Gmsh limitations on narrow cylinders.
if False:
    model.commit_geometry()
    model.mw.set_frequency(1e9)
    model.generate_mesh()
    model.view()

# SET TRUE TO MESH AND VIEW THE GEOMETRY
# This version refines the mesh to avoid intersection errors and label the volumes.
if False:
    # Inspect labeled objects "board_Obj0", "board_Obj1", etc. to identify geometry parts.
    for vol in step.volumes:
        vol.max_meshsize = 1e-3  # smaller element size improves meshing stability
    model.mw.set_frequency(1e9)
    model.generate_mesh()
    model.view(labels=True)

# After identifying each volume from visualization, assign materials accordingly.
trace, diel, ground, conn_diel, conn_block, conn_pin = step.volumes

trace.set_material(em.lib.COPPER)
diel.set_material(em.lib.DIEL_RO4350B)
ground.set_material(em.lib.COPPER)
conn_diel.set_material(em.lib.DIEL_TEFLON)
conn_block.set_material(em.lib.COPPER)
conn_pin.set_material(em.lib.COPPER)

volumes = (trace, diel, ground, conn_diel, conn_block, conn_pin)

# Define a lumped port surface using a small plate geometry at a known connector location.
# These coordinates were found interactively using m.view().
line_w = 1.1*mm
th = (1.01*mm - 0.5*mm)
lumped_port_face = em.geo.Plate((-line_w/2, 35*mm, -1.01*mm), (line_w, 0, 0), (0, 0, th))

# Create an enclosing air box around the full geometry, with given XYZ margins.
air = step.enclose(x_margins=(margin, margin), y_margins=(-1*mm, margin), z_margins=(margin, margin))

# Subtract solid geometries from the air domain to prevent overlapping volumes in meshing.
for vol in volumes:
    air = em.geo.subtract(air, vol, remove_tool=False)

# Commit final geometry setup to the solver.
model.commit_geometry()

# Set local mesh size limits to avoid "PLC Error" during tetrahedralization.
conn_pin.max_meshsize = 0.3*mm
conn_diel.max_meshsize = 0.6*mm
conn_block.max_meshsize = 1*mm

# Define frequency sweep and basic mesh refinement for conductor boundaries.
model.mw.set_frequency_range(0.1e9, 10e9, 41)
model.mesher.set_boundary_size(trace, 2*mm)
model.generate_mesh()

# View mesh to verify structure and ensure no errors before solving.
model.view(plot_mesh=True, volume_mesh=False)

# Identify boundary faces using automatic directional tags: -x, +x, -y, +y, -z, +z.
# Here, the coaxial connector input is on the -y face.
coax_port_face = conn_diel.face('-y')
# Radiation boundary is applied to all outer box faces except the coax connector.
radiation_face = air.faces('-x', '+x', '-z', '+z', '+y')
model.view(selections=[coax_port_face,])

# Define boundary conditions:
#   Port 1: coaxial modal port on the connector side.
#   Radiation: absorbing boundary on air box.
#   Port 2: lumped port representing microstrip connection.
model.mw.bc.ModalPort(coax_port_face, 1, modetype='TEM')
model.mw.bc.AbsorbingBoundary(radiation_face)
model.mw.bc.LumpedPort(lumped_port_face, 2, line_w, th, (0, 0, 1))

# Run the frequency-domain sweep with parallel processing enabled.
data = model.mw.run_sweep(True, 4, frequency_groups=8)

# Extract structured S-parameter data.
g = data.scalar.grid

# Plot S11 and S21 responses to observe connector-to-trace coupling.
plot_sp(g.freq, [g.S(1,1), g.S(2,1)], labels=['S11', 'S21'])

# Extract field data at 2.25 GHz for near-field and far-field visualization.
field = data.field.find(freq=2.25e9)

# Compute far-field pattern on the defined radiation boundary.
ff = field.farfield_3d(radiation_face)

# Setup visualization with geometry, field cutplane, and far-field surface overlay.
d = model.display
d.add_objects(*model.all_geos())
d.add_surf(*field.cutplane(0.5*mm, x=0).scalar('Ez', 'real'), symmetrize=True)
d.add_surf(*ff.surfplot('normE', rmax=10*mm, offset=(0, 25*mm, 0)))
d.show()
